@@ -53,9 +53,37 @@ except ImportError:
             return self.sock.close()
 
 try:
-    import xml.etree.cElementTree as etree
+    import xml.etree.ElementTree as etree
 except ImportError:
-    import cElementTree as etree
+    import elementtree.ElementTree as etree
+
+# Oh the joys of monkeypatching...
+# We need a CDATA element in set_security_msg, but ElementTree doesn't support it
+def CDATA(text=None):
+    element = etree.Element('![CDATA[')
+    element.text = text
+    return element
+
+# Python 2.7 and 3
+if hasattr(etree, '_serialize_xml'):
+    etree._original_serialize_xml = etree._serialize_xml
+    def _serialize_xml(write, elem, *args):
+        if elem.tag == '![CDATA[':
+            write("\n<%s%s]]>\n" % (elem.tag, elem.text))
+            return
+        return etree._original_serialize_xml(write, elem, *args)
+    etree._serialize_xml = etree._serialize['xml'] = _serialize_xml
+# Python 2.5-2.6, and non-stdlib ElementTree
+elif hasattr(etree.ElementTree, '_write'):
+    etree.ElementTree._orig_write = etree.ElementTree._write
+    def _write(self, file, node, encoding, namespaces):
+        if node.tag == '![CDATA[':
+            file.write("\n<![CDATA[%s]]>\n" % node.text.encode(encoding))
+        else:
+            self._orig_write(file, node, encoding, namespaces)
+    etree.ElementTree._write = _write
+else:
+    raise RuntimeError("Don't know how to monkeypatch CDATA support. Please report a bug at https://github.com/seveas/python-hpilo")
 
 # Which protocol to use
 ILO_RAW  = 1
@@ -129,6 +157,8 @@ class Ilo(object):
         self._elements = None
         self._processors = []
         self.ssl_version = ssl.PROTOCOL_TLSv1
+        self.save_response = None
+        self.read_response = None
 
     def __str__(self):
         return "iLO interface of %s" % self.hostname
@@ -146,8 +176,7 @@ class Ilo(object):
     def _request(self, xml, progress=None):
         """Given an ElementTree.Element, serialize it and do the request.
            Returns an ElementTree.Element containing the response"""
-
-        if not self.protocol:
+        if not self.protocol and not self.read_response:
             self._detect_protocol()
 
         # Serialize the XML
@@ -250,6 +279,17 @@ class Ilo(object):
 
     def _get_socket(self):
         """Set up a subprocess or an https connection and do an HTTP/raw socket request"""
+        if self.read_response:
+            class FakeSocket(object):
+                def __init__(self, file):
+                    self.trash = StringIO.StringIO()
+                    self.output = open(file)
+                    self.read = self.output.read
+                    self.write = self.trash.write
+                    self.close = self.output.close
+                shutdown = lambda *args: None
+            return FakeSocket(self.read_response)
+
         if self.protocol == ILO_LOCAL:
             self._debug(1, "Launching hponcfg")
             try:
@@ -261,16 +301,32 @@ class Ilo(object):
             sp.read = sp.stdout.read
             return sp
 
-        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock.settimeout(self.timeout)
-        self._debug(1, "Connecting to %s:%d" % (self.hostname, self.port))
-        try:
-            sock.connect((self.hostname, self.port))
-        except socket.timeout:
-            raise IloCommunicationError("Timeout connecting to %s:%d" % (self.hostname, self.port))
-        except socket.error:
-            e = sys.exc_info()[1]
-            raise IloCommunicationError("Error connecting to %s:%d: %s" % (self.hostname, self.port, str(e)))
+        self._debug(1, "Connecting to %s port %d" % (self.hostname, self.port))
+        err = None
+        for res in socket.getaddrinfo(self.hostname, self.port, 0, socket.SOCK_STREAM):
+            af, socktype, proto, canonname, sa = res
+            sock = None
+            try:
+                sock = socket.socket(af, socktype, proto)
+                sock.settimeout(self.timeout)
+                self._debug(2, "Connecting to %s port %d" % sa[:2])
+                sock.connect(sa)
+            except socket.timeout:
+                if sock is not None:
+                    sock.close()
+                err = IloCommunicationError("Timeout connecting to %s port %d" % (self.hostname, self.port))
+            except socket.error:
+                if sock is not None:
+                    sock.close()
+                e = sys.exc_info()[1]
+                err = IloCommunicationError("Error connecting to %s port %d: %s" % (self.hostname, self.port, str(e)))
+
+        if err is not None:
+            raise err
+
+        if not sock:
+            raise IloCommunicationError("Unable to resolve %s" % self.hostname)
+
         try:
             return ssl.wrap_socket(sock, ssl_version=self.ssl_version)
         except socket.sslerror:
@@ -356,6 +412,10 @@ class Ilo(object):
         elif sock.shutdown:
             sock.shutdown(socket.SHUT_RDWR)
             sock.close()
+        if self.save_response:
+            fd = open(self.save_response, 'a')
+            fd.write(data)
+            fd.close()
 
         # Stript out garbage from hponcfg
         if self.protocol == ILO_LOCAL:
@@ -485,7 +545,12 @@ class Ilo(object):
             if isinstance(retval, list):
                 retval.append(val)
             elif key in retval:
-                retval[key].update(val)
+                if isinstance(retval[key], dict):
+                    retval[key].update(val)
+                elif not isinstance(retval[key], list):
+                    retval[key] = [retval[key], val]
+                else:
+                    retval[key].append(val)
             else:
                 retval[key] = val
         return retval
@@ -643,9 +708,8 @@ class Ilo(object):
         return self._control_tag('USER_INFO', 'ADD_USER', elements=elements,
                 attrib={'USER_LOGIN': user_login, 'USER_NAME': user_name, 'PASSWORD': password})
 
-    @untested
     def ahs_clear_data(self):
-        """FIXME: I have no relevant hardware. Please report sample output"""
+        """Clears Active Health System information log"""
         return self._raw(('RIB_INFO', {'MODE': 'WRITE'}), ('AHS_CLEAR_DATA', {}))
 
     def cert_fqdn(self, use_fqdn):
@@ -653,8 +717,12 @@ class Ilo(object):
         use_fqdn = str({True: 'Yes', False: 'No'}.get(use_fqdn, use_fqdn))
         return self._control_tag('RIB_INFO', 'CERT_FQDN', attrib={'VALUE': use_fqdn})
 
-    def certificate_signing_request(self):
+    def certificate_signing_request(self, country=None, state=None, locality=None, organization=None,
+            organizational_unit=None, common_name=None):
         """Get a certificate signing request from the iLO"""
+        vars = locals()
+        del vars['self']
+        vars = dict([('CSR_' + x.upper(), vars[x]) for x in vars if vars[x]])
         return self._control_tag('RIB_INFO', 'CERTIFICATE_SIGNING_REQUEST')
 
     def clear_ilo_event_log(self):
@@ -684,6 +752,10 @@ class Ilo(object):
         """Delete the specified user from the ilo"""
         return self._control_tag('USER_INFO', 'DELETE_USER', attrib={'USER_LOGIN': user_login})
 
+    def disable_ers(self):
+        """Disable Insight Remote Support functionality and unregister the server"""
+        return self._control_tag('RIB_INFO', 'DISABLE_ERS')
+
     def eject_virtual_floppy(self):
         """Eject the virtual floppy"""
         return self._control_tag('RIB_INFO', 'EJECT_VIRTUAL_FLOPPY')
@@ -693,12 +765,26 @@ class Ilo(object):
         return self._control_tag('RIB_INFO', 'EJECT_VIRTUAL_MEDIA',
                 attrib={"DEVICE": device.upper()})
 
+    def ers_ahs_submit(self, message_id, bb_days):
+        """Submity AHS dat to the insight remote support server"""
+        elements = [
+            etree.Element('MESSAGE_ID', attrib={'VALUE': str(message_id)}),
+            etree.Element('BB_DAYS', attrib={'VALUE': str(bb_days)}),
+        ]
+        return self._control_tag('RIB_INFO', 'TRIGGER_BB_DATA', elements=elements)
+
+    def fips_enable(self):
+        """Enable FIPS standard to enforce AES/3DES encryption, can only be
+           reset with a call to factory_defaults. Resets Administrator password
+           and license key"""
+        return self._control_tag('RIB_INFO', 'FIPS_ENABLE')
+
     def factory_defaults(self):
         """Reset the iLO to factory default settings"""
         return self._control_tag('RIB_INFO', 'FACTORY_DEFAULTS')
 
     def get_ahs_status(self):
-        """Get active health system  logging status"""
+        """Get active health system logging status"""
         return self._info_tag('RIB_INFO', 'GET_AHS_STATUS')
 
     def get_all_users(self):
@@ -718,6 +804,15 @@ class Ilo(object):
             return dict([(x['user_login'], x) for x in data])
         return self._info_tag('USER_INFO', 'GET_ALL_USER_INFO', process=process)
 
+    def get_asset_tag(self):
+        """Gets the server asset tag"""
+        # The absence of an asset tag is communicated in a warning and there
+        # will be *NO* returntag, hence the AttributeError.
+        try:
+            return self._info_tag('SERVER_INFO', 'GET_ASSET_TAG')
+        except AttributeError:
+            return {'asset_tag': None}
+
     def get_cert_subject_info(self):
         """Get ssl certificate subject information"""
         return self._info_tag('RIB_INFO', 'GET_CERT_SUBJECT_INFO', 'CSR_CERT_SETTINGS')
@@ -832,6 +927,14 @@ class Ilo(object):
                 data[tag] = elt.get('VALUE')
         return data
 
+    def get_ers_settings(self):
+        """Get the ERS Insight Remote Support settings"""
+        return self._info_tag('RIB_INFO', 'GET_ERS_SETTINGS')
+
+    def get_fips_status(self):
+        """Is the FIPS-mandated AES/3DESencryption enforcement in place"""
+        return self._info_tag('RIB_INFO', 'GET_FIPS_STATUS')
+
     def get_fw_version(self):
         """Get the iLO firmware version"""
         return self._info_tag('RIB_INFO', 'GET_FW_VERSION')
@@ -877,6 +980,14 @@ class Ilo(object):
         """Get the list of installed languages - broken because iLO returns invalid XML"""
         return self._info_tag('RIB_INFO', 'GET_ALL_LANGUAGES')
 
+    def get_all_licenses(self):
+        """Get a list of all license types and licenses"""
+        def process(data):
+            if not isinstance(data, list):
+                data = data.values()
+            return [dict([(x[0], x[1]['value']) for x in row]) for row in data]
+        return self._info_tag('RIB_INFO', 'GET_ALL_LICENSES', process=process)
+
     def get_network_settings(self):
         """Get the iLO network settings"""
         return self._info_tag('RIB_INFO', 'GET_NETWORK_SETTINGS')
@@ -904,6 +1015,10 @@ class Ilo(object):
             return [x.lower() for x in data]
         return self._info_tag('SERVER_INFO', 'GET_PERSISTENT_BOOT', ('PERSISTENT_BOOT', 'GET_PERSISTENT_BOOT'), process=process)
 
+    def get_pers_mouse_keyboard_enabled(self):
+        """Returns whether persistent mouse and keyboard are enabled"""
+        return self._info_tag('SERVER_INFO', 'GET_PERS_MOUSE_KEYBOARD_ENABLED', process=lambda data: data['persmouse_enabled'])
+
     def get_power_cap(self):
         """Get the power cap setting"""
         return self._info_tag('SERVER_INFO', 'GET_POWER_CAP', process=lambda data: data['power_cap'])
@@ -916,6 +1031,14 @@ class Ilo(object):
         """Get the power and power alert threshold settings"""
         return self._info_tag('SERVER_INFO', 'GET_PWREG')
 
+    def get_rack_settings(self):
+        """Get the rack settings for an iLO"""
+        return self._info_tag('RACK_INFO', 'GET_RACK_SETTINGS')
+
+    def get_security_msg(self):
+        """Retrieve the security message that is displayed on the login screen"""
+        return self._info_tag('RIB_INFO', 'GET_SECURITY_MSG')
+
     def get_server_auto_pwr(self):
         """Get the automatic power on delay setting"""
         return self._info_tag('SERVER_INFO', 'GET_SERVER_AUTO_PWR', process=lambda data: data['server_auto_pwr'])
@@ -940,6 +1063,10 @@ class Ilo(object):
         """Where does the iLO send SNMP traps to and which traps does it send"""
         return self._info_tag('RIB_INFO', 'GET_SNMP_IM_SETTINGS')
 
+    def get_spatial(self):
+        """Get location information"""
+        return self._info_tag('SERVER_INFO', 'GET_SPATIAL', 'SPATIAL')
+
     def get_sso_settings(self):
         """Get the HP SIM Single Sign-On settings"""
         return self._info_tag('SSO_INFO', 'GET_SSO_SETTINGS')
@@ -1004,14 +1131,20 @@ class Ilo(object):
         return self._control_tag('RIB_INFO', 'INSERT_VIRTUAL_MEDIA', attrib={'DEVICE': device.upper(), 'IMAGE_URL': image_url})
 
     def mod_global_settings(self, session_timeout=None, f8_prompt_enabled=None,
-            f8_login_required=None, lock_configuration=None,
+            f8_login_required=None, lock_configuration=None, ilo_funct_enabled=None,
+            serial_cli_status=None, serial_cli_speed=None,
             http_port=None, https_port=None, ssh_port=None, ssh_status=None,
             vmedia_disable=None, virtual_media_port=None, remote_console_port=None,
+            snmp_access_enabled=None, snmp_port=None, snmp_trap_port=None,
+            remote_syslog_enable=None, remote_syslog_server_address=None, remote_syslog_port=None,
+            alertmail_enable=None, alertmail_email_address=None,
+            alertmail_sender_domain=None, alertmail_smtp_server=None,
             min_password=None, enfoce_aes=None, authentication_failure_logging=None,
             rbsu_post_ip=None, remote_console_encryption=None, remote_keyboard_model=None,
             terminal_services_port=None, high_performance_mouse=None,
             shared_console_enable=None, shared_console_port=None,
-            remote_console_acquire=None, brownout_recovery=None):
+            remote_console_acquire=None, brownout_recovery=None,
+            ipmi_dmci_over_lan_enabled=None, vsp_log_enable=None, vsp_software_flow_control=None):
         """Modify iLO global settings, only values that are specified will be changed."""
         vars = dict(locals())
         del vars['self']
@@ -1031,15 +1164,31 @@ class Ilo(object):
             dhcp_sntp_settings=None, sntp_server1=None, sntp_server2=None,
             timezone=None, enclosure_ip_enable=None, web_agent_ip_address=None,
             shared_network_port=None, vlan_enabled=None, vlan_id=None,
-            shared_network_port_vlan=None, shared_network_port_vlan_id=None):
+            shared_network_port_vlan=None, shared_network_port_vlan_id=None, ipv6_address=None,
+            ipv6_static_route_1=None, ipv6_static_route2=None, ipv6_static_route_3=None,
+            ipv6_prim_dns_server=None, ipv6_sec_dns_server=None, ipv6_ter_dns_server=None,
+            ipv6_default_gateway=None, ipv6_preferred_protocol=None, ipv6_addr_autocfg=None,
+            ipv6_reg_ddns_server=None, dhcpv6_dns_server=None, dhcpv6_rapid_commit=None,
+            dhcpv6_stateful_enable=None, dhcpv6_stateless_enable=None, dhcpv6_sntp_settings=None):
         """Configure the network settings for the iLO card"""
         vars = dict(locals())
         del vars['self']
+
+        # For the ipv4 route elements, use "dest gateway" or (dest, gateway) or {'dest': XXX, 'gateway': XXX}
+        # ipv6 routes are ipv6_dest, prefixlen, ipv6_gateway
+        # IPv6 addresses may specify prefixlength as /64 (default 64)
         elements = [etree.Element(x.upper(), VALUE=str({True: 'Yes', False: 'No'}.get(vars[x], vars[x])))
                     for x in vars if vars[x] is not None]
+        for element in elements:
+            if element.tag == 'IPV6_ADDRESS':
+                addr = element.attrib['VALUE']
+                if '/' in addr:
+                    addr, plen = addr.rsplit('/', 1)
+                    element.attrib.update({'VALUE': addr, 'PREFIXLEN': plen})
+                if 'PREFIXLEN' not in element.attrib:
+                    element.attrib['PREFIXLEN'] = '64'
         return self._control_tag('RIB_INFO', 'MOD_NETWORK_SETTINGS', elements=elements)
 
-
     def mod_dir_config(self, dir_authentication_enabled=None,
             dir_local_user_acct=None,dir_server_address=None,
             dir_server_port=None,dir_object_dn=None,dir_object_password=None,
@@ -1083,9 +1232,15 @@ class Ilo(object):
         return self._control_tag('DIR_INFO','MOD_DIR_CONFIG',elements=elements)
 
 
-    def mod_snmp_im_settings(self, web_agent_ip_address=None, snmp_address_1=None,
-            snmp_address_2=None, snmp_address_3=None, os_traps=None,
-            snmp_passthrough_status=None, rib_traps=None, cim_security_mask=None):
+    def mod_snmp_im_settings(self, snmp_access=None, web_agent_ip_address=None,
+            snmp_address_1=None, snmp_address_1_rocommunity=None, snmp_address_1_trapcommunity=None,
+            snmp_address_2=None, snmp_address_2_rocommunity=None, snmp_address_2_trapcommunity=None,
+            snmp_address_3=None, snmp_address_3_rocommunity=None, snmp_address_3_trapcommunity=None,
+            snmp_port=None, snmp_trap_port=None, snmp_v3_engine_id=None, snmp_passthrough_status=None,
+            trap_source_identifier=None, os_traps=None, rib_traps=None, cold_start_trap_broadcast=None,
+            snmp_v1_traps=None, cim_security_mask=None, snmp_sys_location=None, snmp_sys_contact=None,
+            agentless_management_enable=None, snmp_system_role=None, snmp_system_role_detail=None):
+        # FIXME SNMP User profiles
         """Configure the SNMP and Insight Manager integration settings"""
         vars = dict(locals())
         del vars['self']
@@ -1093,6 +1248,25 @@ class Ilo(object):
                     for x in vars if vars[x] is not None]
         return self._control_tag('RIB_INFO', 'MOD_SNMP_IM_SETTINGS', elements=elements)
 
+    def mod_sso_settings(self, trust_mode=None, user_remote_cons_priv=None,
+            user_reset_server_priv=None, user_virtual_media_priv=None,
+            user_config_ilo_priv=None, user_admin_priv=None,
+            operator_login_priv=None, operator_remote_cons_priv=None,
+            operator_reset_server_priv=None, operator_virtual_media_priv=None,
+            operator_config_ilo_priv=None, operator_admin_priv=None,
+            administrator_login_priv=None, administrator_remote_cons_priv=None,
+            administrator_reset_server_priv=None, administrator_virtual_media_priv=None,
+            administrator_config_ilo_priv=None, administrator_admin_priv=None):
+        vars = dict(locals())
+        del vars['self']
+        del vars['trust_mode']
+        elements = []
+        if trust_mode is not None:
+            elements.append(etree.Element('TRUST_MODE', attrib={'VALUE': trust_mode}))
+        vars = [(x.upper().split('_', 1), {True: 'Yes', False: 'No'}.get(vars[x], vars[x])) for x in vars if vars[x]]
+        elements += [etree.Element(x[0][0] + '_ROLE', attrib={x[0][1]: x[1]}) for x in vars]
+        return self._control_tag('SSO_INFO', 'MOD_SSO_SETTINGS', elements=elements)
+
     def mod_user(self, user_login, user_name=None, password=None,
             admin_priv=None, remote_cons_priv=None, reset_server_priv=None,
             virtual_media_priv=None, config_ilo_priv=None):
@@ -1115,6 +1289,18 @@ class Ilo(object):
         """Press the power button"""
         return self._control_tag('SERVER_INFO', 'PRESS_PWR_BTN')
 
+    def profile_apply_get_results(self):
+        """Retrieve the results of the last profile_apply"""
+        return self._info_tag('RIB_INFO', 'PROFILE_APPLY_GET_RESULTS')
+
+    def profile_list(self):
+        """List all profile descriptors"""
+        def process(data):
+            if isinstance(data, dict):
+                return data.values()
+            return data
+        return self._info_tag('RIB_INFO', 'PROFILE_LIST', 'PROFILE_DESC_LIST', process=process)
+
     def hold_pwr_btn(self):
         """Press and hold the power button"""
         return self._control_tag('SERVER_INFO', 'HOLD_PWR_BTN')
@@ -1140,8 +1326,20 @@ class Ilo(object):
         status = {True: 'enable', False: 'disable'}[status]
         return self._control_tag('RIB_INFO', 'SET_AHS_STATUS', attrib={'VALUE': status})
 
+    def set_asset_tag(self, asset_tag):
+        """Set the server asset tag"""
+        return self._control_tag('SERVER_INFO', 'SET_ASSET_TAG', attrib={'VALUE': asset_tag})
+
+    def set_ers_irs_connect(self, ers_destination_url, ers_destination_port):
+        """Connect to an Insight Remote Support server"""
+        elements = [
+            etree.Element('ERS_DESTINATION_URL', attrib={'VALUE': ers_destination_url}),
+            etree.Element('ERS_DESTINATION_PORT', attrib={'VALUE': str(ers_destination_port)}),
+        ]
+        return self._control_tag('RIB_INFO', 'SET_ERS_IRS_CONNECT', elements=elements)
+
     def set_language(self, lang_id):
-        """Set the default language"""
+        """Set the default language. Only EN, JA and ZH are supported"""
         return self._control_tag('RIB_INFO', 'SET_LANGUAGE', attrib={'LANG_ID': lang_id})
 
     def set_host_power(self, host_power=True):
@@ -1154,7 +1352,13 @@ class Ilo(object):
         return self._control_tag('SERVER_INFO', 'SET_HOST_POWER_SAVER', attrib={'HOST_POWER_SAVER': str(host_power_saver)})
 
     def set_one_time_boot(self, device):
-        """Set one time boot device"""
+        """Set one time boot device, device should be one of normal, floppy,
+           cdrom, hdd, usb, rbsu or network. Ilo 4 also supports EMB-MENU
+           (Displays the default boot menu), EMB-ACU (Boots into ACU),
+           EMB-HPSUM-AUTO (Boots HPSUM in automatic update mode), EMB-DIAGS
+           (Launches Insight Diagnostics for Linux in interactive mode) and
+           RBSU (Boots into the system RBSU)"""
+
         return self._control_tag('SERVER_INFO', 'SET_ONE_TIME_BOOT', attrib={'VALUE': device.upper()})
 
     def set_persistent_boot(self, devices):
@@ -1162,6 +1366,11 @@ class Ilo(object):
         elements = [etree.Element('DEVICE', VALUE=x.upper()) for x in devices.split(',')]
         return self._control_tag('SERVER_INFO', 'SET_PERSISTENT_BOOT', elements=elements)
 
+    def set_pers_mouse_keyboard_enabled(self, enabled):
+        """Enable/disable persistent mouse and keyboard"""
+        enabled = {True: 'Yes', False: 'No'}.get(enabled,enabled)
+        return self._control_tag('SERVER_INFO', 'SET_PERS_MOUSE_KEYBOARD_ENABLED', attrib={'VALUE': enabled})
+
     def set_pwreg(self, type, threshold=None, duration=None):
         """Set the power alert threshold"""
         elements = [etree.Element('PWRALERT', TYPE=type)]
@@ -1173,6 +1382,14 @@ class Ilo(object):
         """Set the power cap feature to a specific value"""
         return self._control_tag('SERVER_INFO', 'SET_POWER_CAP', attrib={'POWER_CAP': str(power_cap)})
 
+    def set_security_msg(self, security_msg, security_msg_text=''):
+        """Enables/disables the security message on the iLO login screen and sets its value"""
+        enabled = str({True: 'Yes', False: 'No'}.get(security_msg, security_msg))
+        text = etree.Element('SECURITY_MSG_TEXT')
+        text.append(CDATA(security_msg_text))
+        elements = (etree.Element('SECURITY_MSG', VALUE=enabled), text)
+        return self._control_tag('RIB_INFO', 'SET_SECURITY_MSG', elements=elements)
+
     def set_server_auto_pwr(self, setting):
         """Set the automatic power on delay setting. Valid settings are False,
            True (for minumum delay), 15, 30, 45 60 (for that amount of delay or
@@ -1208,6 +1425,16 @@ class Ilo(object):
         return self._control_tag('RIB_INFO', 'SET_VM_STATUS', attrib={'DEVICE': device.upper()},
                                  elements=elements)
 
+    def trigger_l2_collection(self, message_id):
+        """Initiate an L2 data collection submission to the Insight Remote Support server."""
+        element = etree.Element('MESSAGE_ID', attrib={'value': str(message_id)})
+        return self._control_tag('RIB_INFO', 'TRIGGER_L2_COLLECTION', elements=[element])
+
+    def trigger_test_event(self, message_id):
+        """Trigger a test service event submission to the Insight Remote Support server."""
+        element = etree.Element('MESSAGE_ID', attrib={'value': str(message_id)})
+        return self._control_tag('RIB_INFO', 'TRIGGER_TEST_EVENT', elements=[element])
+
     def uid_control(self, uid="No"):
         """Turn the UID light on ("Yes") or off ("No")"""
         if uid.lower() not in ('yes', 'no'):
@@ -1216,7 +1443,15 @@ class Ilo(object):
 
     def update_rib_firmware(self, filename, progress=None):
         """Upload new RIB firmware, use "latest" as filename to automatically
-        download and use the latest firmware"""
+           download and use the latest firmware. As this function may take a
+           while, you can get progress notifications by passing a callable in
+           the progress parameter. This callable will be called with progress
+           messages. These messages either start with a carriage return ('\\\\r')
+           and an optional ANSI 'clear line' sequence ('\\\\033[K'), or without
+           one. When printing to a terminal, append a newline only to the
+           second type of string. When using the data outside a terminal
+           environment, make sure you strip off the carriage return and ansi
+           sequence"""
         if self.delayed:
             raise IloError("Cannot run firmware update in delayed mode")
 
@@ -1261,6 +1496,10 @@ class Ilo(object):
             self._upload_file(filename, progress_)
             return self._request(root, progress_)[1]
 
+# TODO
+# - All the profile functions page 111-116
+# - sso_server / delete_server
+# - mod_twofactor_settings / import_2factor_cert
 
 ##############################################################################################
 #### All functions below require hardware I don't have access to
@@ -1425,7 +1664,6 @@ class Ilo(object):
                 self.assertTrue('fans' in res)
                 self.assertTrue('fans' in res['health_at_a_glance'])
 
-                return # FIXME - need to implement this first
                 if self.ilo_version >= 3:
                     self.assertTrue('drives' in res)
                 if self.ilo_version >= 4:
